from django.core.exceptions import ValidationError
from django.db import models

from enumeration.const import ResetPeriod
from enumeration.validators import validate_format


class Sequence(models.Model):
    format = models.CharField(max_length=255)
    reset_period = models.CharField(
        ResetPeriod, max_length=30, default=ResetPeriod.NEVER
    )

    def clean(self):
        try:
            validate_format(self.format, self.reset_period)

        except ValidationError as e:
            raise ValidationError({"format": e.messages})


class Counter(models.Model):
    sequence = models.ForeignKey(
        Sequence, related_name="counters", on_delete=models.CASCADE
    )
    position = models.PositiveIntegerField(default=0)
    period = models.DateField(null=True)

    class Meta:
        unique_together = ("sequence", "period")
        constraints = [
            # only one counter for period-naive sequences
            models.UniqueConstraint(
                fields=["sequence"],
                condition=models.Q(period__isnull=True),
                name="unique_counter_for_no_period",
            )
        ]


class Gap(models.Model):
    counter = models.ForeignKey(Counter, related_name="gaps", on_delete=models.CASCADE)
    position = models.PositiveIntegerField()

    unique_together = ("counter", "position")


# class EnumeratedDocumentMixin(models.Model):
#     sequence = models.ForeignKey(Sequence)
#     position = models.PositiveIntegerField(null=True)
#     number = models.CharField(max_length=255, null=True, blank=True)
#
#     class Meta:
#         abstract = True



# Generated by Django 3.2.10 on 2021-12-28 05:38

import django.db.models.deletion

from django.db import migrations
from django.db import models

import enumeration.const


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Counter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("position", models.PositiveIntegerField(default=0)),
                ("period", models.DateField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Sequence",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("format", models.CharField(max_length=255)),
                (
                    "reset_period",
                    models.CharField(
                        default="never",
                        max_length=30,
                        verbose_name=enumeration.const.ResetPeriod,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Gap",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("position", models.PositiveIntegerField()),
                (
                    "counter",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="gaps",
                        to="enumeration.counter",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="counter",
            name="sequence",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="counters",
                to="enumeration.sequence",
            ),
        ),
        migrations.AddConstraint(
            model_name="counter",
            constraint=models.UniqueConstraint(
                condition=models.Q(("period__isnull", True)),
                fields=("sequence",),
                name="unique_counter_for_no_period",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="counter",
            unique_together={("sequence", "period")},
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['enumeration', 'enumeration.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2']

setup_kwargs = {
    'name': 'django-enumeration',
    'version': '1.0.0',
    'description': 'Robust django enumeration sequences',
    'long_description': '============================\nDjango enumeration sequences\n============================\n\n**Consistent, gapless, periodic(optional) sequences with number formatting for accounting documents and such**\n\n\nRequirements\n==============\n\n* Postgres 9.5+\n* Python 3.6+\n* Django 3.2+\n\n\nInstallation\n____________\n\n1. pip install django-enumeration\n\n2. Add "enumeration" to your INSTALLED_APPS setting like this::\n\n    INSTALLED_APPS = [\n        ...\n        \'enumeration\',\n    ]\n\n3. Run `python manage.py migrate` to create the enumeration models.\n\n',
    'author': 'Mārtiņš Šulcs',
    'author_email': 'shulcsm@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2',
}


setup(**setup_kwargs)



import re

from django.core.exceptions import ValidationError

from enumeration.const import ResetPeriod
from enumeration.formatter import TOKENS


POSITION = re.compile(TOKENS["POSITION"])
YEAR = re.compile(TOKENS["YEAR"])
MONTH = re.compile(TOKENS["MONTH"])
DAY = re.compile(TOKENS["DAY"])


def validate_format(format_string, period: ResetPeriod):
    if not POSITION.search(format_string):
        raise ValidationError('Number placeholder "#{1,10}" is required.')

    if period in [ResetPeriod.YEARLY, ResetPeriod.MONTHLY, ResetPeriod.DAILY]:
        if not YEAR.search(format_string):
            raise ValidationError(
                f'Year placeholder "Y{{2,4}}" is required for {period.value} period.'
            )

    if period in [ResetPeriod.MONTHLY, ResetPeriod.DAILY]:
        if not MONTH.search(format_string):
            raise ValidationError(
                f'Month placeholder "M{{1,2}}" is required for {period.value} period.'
            )

    if period == ResetPeriod.DAILY:
        if not DAY.search(format_string):
            raise ValidationError(
                f'Day placeholder "D{{1,2}}" is required for {period.value} period.'
            )



import re


TOKENS = {
    "LITERAL": r"{{.*?}}",
    "CONTEXT": r"{\w+}",
    "POSITION": r"#{1,10}",
    "YEAR": r"Y{2,4}",
    "MONTH": r"M{1,2}",
    "DAY": r"D{1,2}",
    "CHAR": r"[\x00-\x7F]",
    "WS": r"\s+",
}

TOKEN_RE = re.compile("|".join("(?P<%s>%s)" % pair for pair in TOKENS.items()))


def format_number(format_string: str, position: int, **context) -> str:
    parts = []

    for match in TOKEN_RE.finditer(format_string):
        token = match.lastgroup
        assert token

        value = match.group(token)

        if token in ["CHAR", "WS"]:
            parts.append(value)

        elif token == "POSITION":
            parts.append(str(position).rjust(len(value), "0"))

        elif token == "YEAR":
            parts.append(str(context["date"].year)[-len(value) :])

        elif token == "MONTH":
            parts.append(str(context["date"].month).rjust(len(value), "0"))

        elif token == "DAY":
            parts.append(str(context["date"].day).rjust(len(value), "0"))

        elif token == "LITERAL":
            parts.append(value.strip("{}"))

        elif token == "CONTEXT":
            parts.append(context[value.strip("{}")])

    return "".join(parts)
